# -*- coding: utf-8 -*-
"""Module for reading configuration from conffile."""
import io
# Config parser for reading the interface settings
import configparser
# Global settings for rpi2caster
from .global_settings import CONFIG_PATH
# Custom exceptions
from .exceptions import NotConfigured, ConfigFileUnavailable
# Package constants
from .constants import DEFAULT_CONFIG_PATHS, TRUE_ALIASES, FALSE_ALIASES


def initialize_config():
    """Initializes config.

    Looks for the configuration file, tests if it's readable.
    Throws an exception if errors occur.
    """
    config_paths = [CONFIG_PATH] + DEFAULT_CONFIG_PATHS
    for path in config_paths:
        try:
            with io.open(path, 'r'):
                cfg = configparser.SafeConfigParser()
                cfg.read(path)
                return cfg
        except (IOError, FileNotFoundError):
            print('Configuration file at %s failed' % path)
            continue
    # No config file specified can be accessed
    raise ConfigFileUnavailable('Cannot access any config file')


def get_config(section_name, option_name):
    """get_config:

    Gets a value for a given parameter from a given name.
    Returns:
      -int - if the value is numeric,
      -boolean - if the value was in one of the true or false aliases,
      -string otherwise,
      -None if there is no option or section of that name.
    """
    cfg = initialize_config()
    try:
        # Return an integer value whenever we can
        return int(cfg.get(section_name, option_name))
    except (ValueError, TypeError):
        # Get the value and decide what to do with it
        value = cfg.get(section_name, option_name)
        return determine(value)
    except (configparser.NoSectionError, configparser.NoOptionError):
        # If section or option is not configured, raise an exception
        raise NotConfigured


def determine(value):
    """Checks if the string means a hexdigit, None, True or False"""
    # Convert to lowercase:
    value = value.lower()
    try:
        if value.startswith('0x'):
            # Value is a hexstring: 0x or 0X
            return int(value, 16)
        elif value in ['none', 'null']:
            # Value was specified to be None or null
            return None
        elif value in TRUE_ALIASES:
            # 1, on, true, yes
            return True
        elif value in FALSE_ALIASES:
            # 0, off, false, no
            return False
        else:
            # Return an integer
            return int(value)
    except (ValueError, TypeError, AttributeError):
        # Do nothing if the value has no method we tried
        pass
    # Return the raw value - a list, a string, None etc.
    return value


def section_not_found(section_name):
    """section_not_found:

    Indicates whether there's no section with a given name."""
    cfg = initialize_config()
    return not cfg.has_section(section_name)



# -*- coding: utf-8 -*-
"""Web-based CLI for rpi2caster"""

import os.path
import tornado.web
import tornado.ioloop
# This demo requires tornado_xstatic and XStatic-term.js
import tornado_xstatic

import terminado
STATIC_DIR = os.path.join(os.path.dirname(terminado.__file__), "_static")


class TerminalPageHandler(tornado.web.RequestHandler):
    def get(self):
        return self.render("/var/www/rpi2caster/termpage.html",
                           static=self.static_url,
                           xstatic=self.application.settings['xstatic_url'],
                           ws_url_path="/websocket")

if __name__ == '__main__':
    term_manager = terminado.SingleTermManager(shell_command=['rpi2caster'])
    handlers = [
                (r"/websocket", terminado.TermSocket,
                 {'term_manager': term_manager}),
                (r"/", TerminalPageHandler),
                (r"/xstatic/(.*)", tornado_xstatic.XStaticFileHandler,
                 {'allowed_modules': ['termjs']})
               ]
    xstatic_url = tornado_xstatic.url_maker('/xstatic/')
    app = tornado.web.Application(handlers, static_path=STATIC_DIR,
                                  xstatic_url=xstatic_url)
    # Serve at http://localhost:8765/ N.B. Leaving out 'localhost' here will
    # work, but it will listen on the public network interface as well.
    # Given what terminado does, that would be rather a security hole.
    app.listen(8000)
    try:
        tornado.ioloop.IOLoop.instance().start()
    finally:
        term_manager.shutdown()



# -*- coding: utf-8 -*-
"""output_driver_wiringpi - wiringPi2-based output driver for MCP23017"""

# WiringPi2 Python bindings: essential for controlling the MCP23017!
try:
    import wiringpi
except ImportError:
    print('You must install wiringpi!')

# Import mockup output driver from monotype
from .monotype import SimulationOutput, SIGNALS
from .exceptions import WrongConfiguration
from .global_settings import MCP0, MCP1


class Singleton(type):
    """Make only one object"""
    instance = None

    def __call__(cls, *args, **kw):
        if not cls.instance:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


class WiringPiOutputDriver(SimulationOutput):
    """A 32-channel control interface based on two MCP23017 chips"""
    pin_base = 65
    __metaclass__ = Singleton

    def __init__(self, mcp0_address=MCP0, mcp1_address=MCP1, sig_arr=SIGNALS):
        super().__init__(sig_arr=sig_arr)
        self.name = 'MCP23017 driver using wiringPi2-Python library'
        # Set up an output interface on two MCP23017 chips
        pin_base = WiringPiOutputDriver.pin_base
        wiringpi.mcp23017Setup(pin_base, mcp0_address)
        wiringpi.mcp23017Setup(pin_base + 16, mcp1_address)
        pins = [x for x in range(pin_base, pin_base+32)]
        # Update the pin base for next instances
        WiringPiOutputDriver.pin_base += 32
        self.pin_numbers = dict(zip(self.signals_arrangement, pins))
        # Set all I/O lines on MCP23017s as outputs - mode=1
        for pin in self.pin_numbers.values():
            wiringpi.pinMode(pin, 1)

    def one_on(self, sig):
        """Looks a signal up in arrangement and turns it on"""
        try:
            wiringpi.digitalWrite(self.pin_numbers[sig], 1)
        except KeyError:
            msg = ('Signal %s not defined! Signals arrangement: %s\n'
                   % (sig, self.signals_arrangement))
            raise WrongConfiguration(msg)

    def one_off(self, sig):
        """Looks a signal up in arrangement and turns it off"""
        try:
            wiringpi.digitalWrite(self.pin_numbers[sig], 0)
        except KeyError:
            msg = ('Signal %s not defined! Signals arrangement: %s\n'
                   % (sig, self.signals_arrangement))
            raise WrongConfiguration(msg)



# -*- coding: utf-8 -*-
"""
typesetting_functions:

Contains functions used for calculating line length, justification,
setting wedge positions, breaking the line etc.
"""
from collections import deque
from . import exceptions as e
from .justification import Box, Glue, Penalty, ObjectList
from .measure import Measure
from .global_settings import UI
from .matrix_data import Diecase, Matrix

STYLES = {'^00': 'r', '^01': 'b', '^02': 'i',
          '^03': 's', '^04': 'l', '^05': 'u',
          '^RR': 'r', '^BB': 'b', '^II': 'i',
          '^SS': 's', '^LL': 'l', '^UU': 'u'}
ALIGNMENT = {'^CR': 'left', '^CC': 'center', '^CL': 'right', '^CF': 'both'}


class InputData(object):
    """Gets the input text, parses it, generates a sequence of characters
    or commands"""
    def __init__(self, text, context):
        self.context = context
        self.text = text

    def parse_input(self, input_text):
        """Generates a sequence of characters from the input text.
        For each character, this function predicts what two next characters
        and one next character are."""
        # This variable will prevent yielding a number of subsequent chars
        # after a ligature or command has been found and yielded.
        skip_steps = 0
        # Default style is roman
        style = 'r'
        # Characters which will be skipped
        ignored = ('\n',)
        # Determine the length of character combinations parsed
        max_len = max((len(mat.char) for mat in self.context.diecase) +
                      (len(x) for x in ignored) +
                      (len(x) for x in STYLES) +
                      (len(x) for x in ALIGNMENT))
        # Build a dict of matrices
        mats = {(mat.char, mat.styles): mat for mat in self.context.diecase}
        # What if char in text not present in diecase? Hmmm...
        for index, _ in enumerate(input_text):
            if skip_steps:
                # Skip the characters to be skipped
                skip_steps -= 1
                continue
            for i in range(max_len, 0, -1):
                # Start from longest, end with shortest
                try:
                    char = input_text[index:index+i]
                    skip_steps = i - 1
                    if char in ignored:
                        continue
                    elif char in ALIGNMENT:
                        alignment = ALIGNMENT.get(char, 'both')
                        yield Justify(alignment)
                    elif char in STYLES:
                        # Change the current style
                        style = STYLES.get(char, 'r')
                    else:
                        # Try to look it up in spaces
                        yield mats.get((char, style))
                        # End on first (largest) combination found
                        break
                except (TypeError, AttributeError, ValueError):
                    pass


class Justify(object):
    """Justification class"""
    def __init__(self, method='both'):
        pass


class Line(deque):
    """A line of type, justified to the left, center,right or both."""
    def __init__(self, length):
        super().__init__()
        self.length = length

    def __add__(self, new):
        if isinstance(new, Matrix):
            self.append(new)
        elif isinstance(new, (int, float)):
            self.length += new
        elif isinstance(new, (tuple, list, iter)):
            self.extend([x for x in new])

    def __radd__(self, new):
        if isinstance(new, Matrix):
            self.appendleft(new)
        elif isinstance(new, (int, float)):
            self.length += new
        elif isinstance(new, (tuple, list, iter)):
            self.extendleft([x for x in new])

    def render(self):
        """Renders a line into a series of Matrix objects with appropriate
        widths, and justified spaces."""
        # TODO
        pass

    def _align_left(self):
        """Aligns the previous chunk to the left."""
        pass

    def _align_right(self):
        """Aligns the previous chunk to the right."""
        pass

    def _align_center(self):
        """Aligns the previous chunk to the center."""
        pass

    def _align_both(self):
        """Aligns the previous chunk to both edges and ends the line."""
        pass


class Paragraph(object):
    """A page is broken into a series of paragraphs."""
    def __init__(self, width, style, indent, justification):
        self.width = width
        self.style = style
        self.indent = indent
        self.justification = justification

    def __iter__(self):
        return (x for x in self.nodes)

    def render(self):
        """Render all lines into combinations of Monotype codes"""
        pass

    @property
    def nodes(self):
        """Nodes are all characters, including ligatures, spaces
        and line-breaks"""
        return (node for line in self.lines for node in line)

    @property
    def lines(self):
        """Paragraph's lines"""
        return (line for line in self.lines)


class Page(object):
    """A page is a collection of paragraphs"""
    def __init__(self, width, height):
        self.width = width
        self.height = height


class GalleyBuilder(object):
    """Builds a galley from input sequence"""
    def __init__(self, source, diecase=None, measure=Measure(manual=False)):
        self.source = (x for x in source)
        self.points = measure.points
        self.diecase = diecase or Diecase()
        self.cooldown = False
        self.mould_heatup = True
        self.quad_padding = 1

    def preheat_mould(self):
        """Appends two lines of em-quads at the end"""
        chunk = []
        if self.mould_heatup:
            quad = self.diecase.decode_matrix('O15')
            quad_qty = int(self.points // quad.points)
            chunk = ['%s preheat' % quad] * quad_qty
            comment = 'Casting quads for mould heatup'
            chunk.extend(double_justification(comment=comment))
        return chunk * 2

    def make_ribbon(self):
        """Instantiates a Ribbon() object from whatever we've generated"""
        pass

    def build_galley(self):
        """Builds a line of characters from source"""
        def decode_mat(mat):
            """Gets the mat's parameters and stores them
            to avoid recalculation"""
            parameters = {}
            if mat:
                parameters['wedges'] = mat.wedge_positions()
                parameters['points'] = mat.points
                parameters['code'] = str(mat)
                parameters['lowspace'] = mat.islowspace()
            return parameters

        def start_line():
            """Starts a new line"""
            nonlocal points_left, queue
            points_left = self.points - 2 * self.quad_padding * quad['points']
            quads = [quad['code'] + ' quad padding'] * self.quad_padding
            queue.extend(quads)

        def build_line():
            """Puts the matrix in the queue, changing the justification
            wedges if needed, and adding a space for cooldown, if needed."""
            # Declare variables in non-local scope to preserve them
            # after the function exits
            nonlocal queue, points_left, working_mat, current_wedges
            # Take a mat from stash if there is any
            working_mat = working_mat or decode_mat(next(self.source, None))
            # Try to add another character to the line
            # Empty mat = end of line, start filling
            if points_left > working_mat.get('points', 1000):
                # Store wedge positions
                new_wedges = working_mat.get('wedges', (3, 8))
                # Wedges change? Drop in some single justification
                # (not needed if wedge positions were 3, 8)
                if current_wedges != new_wedges:
                    if current_wedges and current_wedges != (3, 8):
                        queue.extend(single_justification(current_wedges))
                    current_wedges = new_wedges
                # Add the mat
                queue.append(working_mat['code'])
                points_left -= working_mat['points']
                # We need to know what comes next
                working_mat = decode_mat(next(self.source, None))
                if working_mat:
                    next_points = space['points'] + working_mat['points']
                    space_needed = (points_left > next_points and not
                                    working_mat.get('lowspace', True))
                    if self.cooldown and space_needed:
                        # Add a space for matrix cooldown
                        queue.append(space['code'] + ' for cooldown')
                        points_left -= space['points']
                    # Exit and loop further
                    return
            # Finish the line
            var_sp = self.diecase.decode_matrix('G1')
            wedges = current_wedges
            current_wedges = None
            while points_left > quad['points']:
                # Coarse fill with quads
                queue.append(quad['code'] + ' coarse filling line')
                points_left -= quad['points']
            while points_left > space['points'] * 2:
                # Fine fill with fixed spaces
                queue.append(space['code'] + ' fine filling line')
                points_left -= space['points']
            if points_left >= var_sp.get_min_points():
                # Put an adjustable space if possible to keep lines equal
                if wedges:
                    queue.extend(single_justification(wedges))
                var_sp.points = points_left
                queue.append(str(var_sp))
                wedges = var_sp.wedge_positions()
            # Always cast as many quads as needed, then put the line out
            queue.extend([quad['code'] + ' quad padding'] * self.quad_padding)
            queue.extend(double_justification(wedges or (3, 8)))
            points_left = 0

        # Store the code and wedge positions to speed up the process
        space = decode_mat(self.diecase.decode_matrix('G1'))
        quad = decode_mat(self.diecase.decode_matrix('O15'))
        working_mat = None
        current_wedges = None
        queue, points_left = end_casting(), 0
        # Build the whole galley, line by line
        while working_mat != {}:
            start_line()
            while points_left > 0:
                build_line()
        return queue + self.preheat_mould()


def single_justification(wedge_positions=(3, 8),
                         comment='Single justification'):
    """Add 0075 + pos_0075, then 0005 + pos_0005"""
    (pos_0075, pos_0005) = wedge_positions
    return pump_start(pos_0075, comment) + pump_stop(pos_0005)


def double_justification(wedge_positions=(3, 8),
                         comment='Double justification'):
    """Add 0075 + pos_0075, then 0005-0075 + pos_0005"""
    (pos_0075, pos_0005) = wedge_positions
    return pump_start(pos_0075, comment) + galley_trip(pos_0005)


def galley_trip(pos_0005=8, comment='Line to the galley'):
    """Put the line to the galley"""
    attach = comment and (' // ' + comment) or ''
    return ['NKJS 0075 0005 %s%s' % (pos_0005, attach)]


def pump_start(pos_0075=3, comment='Starting the pump'):
    """Start the pump and set 0075 wedge"""
    attach = comment and (' // ' + comment) or ''
    return ['NKS 0075 %s%s' % (pos_0075, attach)]


def pump_stop(pos_0005=8, comment='Stopping the pump'):
    """Stop the pump"""
    attach = comment and (' // ' + comment) or ''
    return ['NJS 0005 %s%s' % (pos_0005, attach)]


def end_casting():
    """Alias for ending the casting job"""
    return (pump_stop(comment='End casting') +
            galley_trip(comment='Last line out'))


def high_or_low_space():
    """Chooses high or low space"""
    spaces = {True: '_', False: ' '}
    high_or_low = UI.confirm('High space?', default=False)
    return spaces[high_or_low]



# -*- coding: utf-8 -*-
"""Global settings for rpi2caster"""
from click import get_app_dir
from . import text_ui as UI

# Conffile path
APPDIR = get_app_dir('rpi2caster', force_posix=True, roaming=True)
CONFIG_PATH = APPDIR + '/rpi2caster.conf'
# SQLite3 database path
DATABASE_PATH = APPDIR + '/rpi2caster.db'
# Display debug data
UI.debug_info('Application data directory: %s' % APPDIR)
UI.debug_info('User config file: %s' % CONFIG_PATH)
UI.debug_info('User database file: %s' % DATABASE_PATH)
# Comment symbols for parsing
COMMENT_SYMBOLS = ['**', '*', '//', '##', '#']
# GPIO settings
# Inputs for the casting interface
EMERGENCY_STOP_GPIO = 22
SENSOR_GPIO = 17
# Default hardware interface output data
MCP0 = 0x20
MCP1 = 0x21
PIN_BASE = 65



# -*- coding: utf-8 -*-
"""rpi2caster - top-level script for starting the components"""
from os import system
from sys import argv
import argparse
from . import exceptions as e
from .text_ui import confirm, menu
from .matrix_data import diecase_operations, list_diecases


def cast(args):
    """Casting on an actual caster or simulation"""
    from . import casting_session
    session = casting_session.Casting(ribbon_file=args.ribbon_file,
                                      ribbon_id=args.ribbon_id,
                                      diecase_id=args.diecase_id,
                                      wedge_name=args.wedge_name)
    session.caster.mode.simulation = args.simulation
    session.caster.mode.punching = args.punching
    casting_session.UI.DEBUG_MODE = args.debug
    # Skip menu if casting directly or testing
    if args.direct:
        session.cast_composition()
    elif args.testing:
        session.diagnostics_submenu()
    else:
        session.main_menu()


def composition(args):
    """Text to ribbon translation and justification"""
    from . import typesetting_session
    session = typesetting_session.Typesetting(manual_mode=args.manual_mode,
                                              text_file=args.text_file,
                                              ribbon_file=args.ribbon_file,
                                              diecase_id=args.diecase_id)
    typesetting_session.UI.DEBUG_MODE = args.debug
    session.main_menu()


def update(args):
    """Updates the software"""
    # Upgrade routine
    dev_prompt = 'Testing version (newest features, but unstable)? '
    if confirm('Update the software?', default=False):
        use_dev_version = args.unstable or confirm(dev_prompt, default=False)
        pre = use_dev_version and '--pre' or ''
        print('You may be asked for the admin password...')
        system('sudo pip3 install %s --upgrade rpi2caster' % pre)


def inventory(args):
    """Inventory management - diecase manipulation etc."""
    if args.list_diecases:
        list_diecases()
    else:
        diecase_operations()


def toggle_punching(args):
    """Switch between punching and casting modes"""
    args.punching = not args.punching


def toggle_simulation(args):
    """Switch between simulation and casting/punching modes"""
    args.simulation = not args.simulation


def main_menu(args):
    """Main menu - choose the module"""
    def exit_program(*_):
        """Breaks out of menu"""
        nonlocal finished
        finished = True

    header = ('rpi2caster - computer aided type casting for Monotype '
              'composition / type & rule casters.'
              '\n\nMain menu:\n')
    finished = False
    while not finished:
        options = [(exit_program, 'Exit',
                    'Exits the rpi2caster suite'),
                   (cast, {True: 'Punch ribbon...',
                           False: 'Cast type...'}[args.punching],
                    {True: 'Punch a ribbon with a keyboard\'s paper tower',
                     False: ('Cast composition, sorts or spaces; '
                             'test the machine')}[args.punching]),
                   (composition, 'Typesetting...',
                    'Compose text for casting'),
                   (inventory, 'Diecase manipulation...',
                    'Add, display, edit or remove matrix case definitions'),
                   (update, 'Update the program',
                    'Check whether new version is available and update'),
                   (toggle_punching, 'Switch to %s mode'
                    % {True: 'casting',
                       False: 'ribbon punching'}[args.punching],
                    ('The casting program has different functionality '
                     'in casting and punching modes.')),
                   (toggle_simulation, 'Switch to %s mode'
                    % {True: 'casting or ribbon punching',
                       False: 'simulation'}[args.simulation],
                    {True: 'Use a real machine',
                     False: 'Use a mockup for testing'}[args.simulation])]
        try:
            menu(options, header=header, footer='')(args)
        except (e.ReturnToMenu, e.MenuLevelUp):
            pass
        except (KeyboardInterrupt, EOFError):
            finished = True


def main():
    """Main function

    Parse input options and decide which to run"""
    # Help description and epilogue i.e. what you see at the bottom
    desc = ('Starting rpi2caster without arguments will open the main menu, '
            'where you can choose what to do (casting, inventory management, '
            'typesetting), toggle simulation or perforation modes.')
    epi = ('Enter "%s [command] -h" for detailed help about its options. '
           'Typesetting is not ready yet.' % argv[0])
    # Initialize the main arguments parser
    main_parser = argparse.ArgumentParser(description=desc, epilog=epi)
    # Set default values for all options globally
    main_parser.set_defaults(job=main_menu, debug=False, ribbon_file='',
                             ribbon_id='', source=None, simulation=False,
                             punching=False, unstable=False,
                             manual_mode=False, list_diecases=False,
                             diecase_id='', wedge_name='',
                             direct=False, testing=False)
    #
    # Define commands
    #
    # Subparsers for jobs: casting, inventory management, typesetting
    jobs = main_parser.add_subparsers(title='Commands',
                                      description='Choose what you want to do',
                                      help='description:')
    #
    # Casting subparser
    #
    cast_parser = jobs.add_parser('cast', aliases=['c'],
                                  help=('Casting with a Monotype caster '
                                        'or mockup caster for testing'))
    # Debug mode
    cast_parser.add_argument('-d', '--debug', help='debug mode',
                             action="store_true")
    # Choose specific diecase
    cast_parser.add_argument('-m', '--diecase', metavar='ID',
                             dest='diecase_id', help='diecase ID for casting')
    # Punch ribbon: uses different sensor, always adds O+15 to combinations
    cast_parser.add_argument('-p', '--punch', action='store_true',
                             dest='punching',
                             help='ribbon punching (perforation) mode')
    # Simulation mode: casting/punching without the actual caster/interface
    cast_parser.add_argument('-s', '--simulation', action='store_true',
                             help='simulate casting instead of real casting')
    test_or_cast = cast_parser.add_mutually_exclusive_group()
    # Starts in the diagnostics submenu of the casting program
    test_or_cast.add_argument('-T', '--test', action='store_true',
                              dest='testing',
                              help='caster / interface diagnostics')
    # Allows to start casting right away without entering menu
    test_or_cast.add_argument('-D', '--direct', action="store_true",
                              help='direct casting - no menu',)
    # Choose specific wedge
    cast_parser.add_argument('-w', '--wedge', metavar='W', dest='wedge_name',
                             help='wedge to use: [s]series-set_width[e]')
    # Ribbon ID for choosing from database
    cast_parser.add_argument('-R', '--ribbon_id', metavar='R',
                             help='ribbon ID to choose from database')
    # Ribbon - input file specification
    cast_parser.add_argument('ribbon_file', metavar='ribbon', nargs='?',
                             help='ribbon file name')
    cast_parser.set_defaults(job=cast)
    #
    # Software update subparser
    #
    upd_parser = jobs.add_parser('update', aliases=['u', 'upd'],
                                 help='Update the software')
    # Update to unstable version
    upd_parser.add_argument('-t', '--testing', action='store_true',
                            help='use testing rather than stable version')
    upd_parser.set_defaults(job=update)
    #
    # Composition (typesetting) program subparser
    #
    comp_parser = jobs.add_parser('translate', aliases=['t', 'set'],
                                  help='Typesetting program')
    # Manual mode option - more control for user
    comp_parser.add_argument('-m', '--manual', help='use manual mode',
                             dest='manual_mode', action='store_true')
    # Choose diecase layout
    comp_parser.add_argument('-D', '--diecase', dest='diecase_id',
                             help='diecase ID for typesetting',
                             metavar='ID')
    # Debug mode
    comp_parser.add_argument('-d', '--debug', help='debug mode',
                             action="store_true")
    # Input filename option
    comp_parser.add_argument('text_file',
                             help='source (text) file to translate',
                             metavar='text_file', nargs='?',
                             type=argparse.FileType('r', encoding='UTF-8'))
    # Output filename option
    comp_parser.add_argument('ribbon_file', help='ribbon file to generate',
                             metavar='ribbon_file', nargs='?',
                             type=argparse.FileType('w', encoding='UTF-8'))
    # Default action
    comp_parser.set_defaults(job=composition)
    #
    # Inventory subparser
    #
    inv_parser = jobs.add_parser('inventory', aliases=['i', 'inv'],
                                 help='Matrix case management')
    # List the diecases and exit
    inv_parser.add_argument('-l', '--list_diecases', action='store_true',
                            help='list all diecases and finish')
    inv_parser.set_defaults(job=inventory)
    # Parsers are defined - get the args...
    args = main_parser.parse_args()
    try:
        args.job(args)
    except (e.ReturnToMenu, e.MenuLevelUp):
        print('Goodbye!')
    except (KeyboardInterrupt, EOFError):
        print('\nInterrupted by user.')
    except AttributeError:
        if args.debug:
            raise
        else:
            main_parser.print_help()
    finally:
        print('Goodbye!')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Parser functions.

This module contains file- and line-parsing functions for the casting program.
"""
import io
from . import constants
# Check if alternate comment symbols are configured
try:
    from .global_settings import COMMENT_SYMBOLS
except ImportError:
    COMMENT_SYMBOLS = constants.COMMENT_SYMBOLS


def read_file(filename):
    """Returns file contents (if file is readable), or False otherwise.
    Strips whitespace and removes empty lines."""
    # Open a file with signals, test if it's readable and return its contents
    try:
        with io.open(filename, 'r') as input_file:
            return [l.strip() for l in input_file if l.strip()]
    except (IOError, FileNotFoundError):
        return False


def parse_record(input_data):
    """Parses an input string, and returns a list with two elements:
    -the Monotype signals (unprocessed),
    -any comments delimited by symbols from COMMENT_SYMBOLS list.
    We need to work on strings. Convert any lists, integers etc.

    Looks for any comment symbols defined here - **, *, ##, #, // etc.
    splits the line at it and saves the comment to return it later on.
    If it's an inline comment (placed after Monotype code combination),
    this combination will be returned for casting.

    If a line in file contains a comment only, returns no combination.

    In case of column O and row 15 (no signals fed to machine), we still
    need to have the signals listed explicitly in the input sequence.
    The signals_parser will later remove them and convert to O15.

    Example:
    O15 //comment      <-- casts from O+15 matrix, displays comment
                       <-- nothing to do
    //comment          <-- displays comment, no casting
    0005 5 //comment   <-- sets 0005 justification wedge to 5,
                           turns pump off, displays comment.
    """
    # Should work both on lists and strings
    input_data = ''.join([str(x) for x in input_data])
    # Assume we don't have a comment...
    raw_signals = input_data
    comment = ''
    # ...then look for comment symbols and parse them:
    for symbol in COMMENT_SYMBOLS:
        if symbol in input_data:
            # Split on the first encountered symbol
            [raw_signals, comment] = input_data.split(symbol, 1)
            break
    # Parse the signals and return a list with processed signals and comment
    return [parse_signals(raw_signals), comment.strip()]


def parse_signals(input_signals):
    """Parses a string with Monotype signals on input.
    Skips all but the "useful" signals: A...O, 1...15, 0005, S, 0075.
    Outputs a list of signals to be processed by the machine control routines.
    Filter out all non-alphanumeric characters and whitespace.
    Convert to uppercase.
    """
    col_sigs = 'ABCDEFGHIJKLMNS'
    signals = (str(x).upper().strip() for x in input_signals)
    signals = (x for x in signals if x.isdigit() or x in col_sigs + 'O')
    signals = ''.join(signals)
    # Find O+15 or 16th row
    o15_found = 'O15' in signals or 'O' in signals or '15' in signals
    row_16_found = '16' in signals
    # Build a list of justification signals
    justification = [sig for sig in ['0005', '0075'] if sig in signals]
    # Remove these signals from the input string
    for sig in justification + [str(i) for i in range(100, 14, -1)]:
        # We operate on a string, so cannot remove the item...
        signals = signals.replace(sig, '')
    # From remaining numbers, determine row numbers.
    # The highest number will be removed from the raw_signals to avoid
    # erroneously adding its digits as signals.
    rows = row_16_found and ['16'] or []
    for row in range(14, 0, -1):
        if str(row) in signals:
            rows.append(str(row))
            signals = signals.replace(str(row), '')
    # Columns + S justification signal
    columns = [s for s in col_sigs if s in signals]
    # Return a list containing all signals and O15 if needed, if the input
    # string or list contained any useful combinations
    return justification + columns + rows + ['O15'] * o15_found or []


def count_lines_and_chars(contents):
    """Count newlines and characters+spaces in ribbon file.
    This is usually called when pre-processing the file for casting.
    """
    all_lines = 0
    all_chars = 0
    for line in contents:
        # Strip comments
        signals = parse_record(line)[0]
        # Parse the signals part of the line
        if check_character(signals):
            all_chars += 1
        elif check_newline(signals):
            all_lines += 1
    # -1 lines because of the starting galley trip / double justification code
    # Cannot be negative
    all_lines = max(0, all_lines - 1)
    return [all_lines, all_chars]


def count_combinations(contents):
    """Count all combinations in ribbon file.
    This is usually called when pre-processing the file for punching."""
    all_combinations = 0
    for line in contents:
        # Get signals list. Increase counter if we have signals.
        if parse_record(line)[0]:
            all_combinations += 1
    # Return the number
    return all_combinations


def stop_comes_first(contents):
    """Detects ribbon direction so that we can use the ribbons generated
    with different software and still cast them in correct order.
    This function loops over the ribbon to test if the stop sequence
    (NJ or 0005) comes before the newline sequence (NKJ or 0075+0005).
    If so, returns True (the ribbon needs to be rewound for casting).
    Otherwise, False."""
    start_found = False
    for line in contents:
        # Get the signals part of a line, discard the comment
        signals, _ = parse_record(line)
        # Toggle this to True if newline combinations are found
        start_found = start_found or check_newline(signals)
        # Determine the result the first time pump stop combination is found
        if check_pump_stop(signals) and not start_found:
            return True
        elif start_found:
            return False


def get_column(signals):
    """Gets the lowest column number from the signals list."""
    decoded_signals = ({'N', 'I'}.issubset(signals) and 'NI' or
                       {'N', 'L'}.issubset(signals) and 'NL' or signals)
    generator = (x for x in constants.COLUMNS_17 if x in decoded_signals)
    try:
        return next(generator)
    except StopIteration:
        return 'O'


def get_row(signals):
    """Gets the lowest row number from the signal list. Returns int."""
    generator = (row for row in range(17) if str(row) in signals)
    try:
        return next(generator)
    except StopIteration:
        return 15


def check_0075(signals):
    """Checks if signals contain 0075 or NK."""
    return '0075' in signals or {'N', 'K'}.issubset(signals)


def check_0005(signals):
    """Checks if signals contain 0005 or NJ."""
    return '0005' in signals or {'N', 'J'}.issubset(signals)


def check_newline(signals):
    """Checks if the newline (0005, 0075 or NKJ) is present in combination.
    This is called for each new line when parsing the ribbon file when casting.
    """
    return check_0075(signals) and check_0005(signals)


def check_pump_stop(signals):
    """Checks if the signals will stop the pump"""
    return check_0005(signals) and not check_0075(signals)


def check_pump_start(signals):
    """Checks if the signals will stop the pump"""
    return check_0075(signals)


def check_character(signals):
    """Check if this is a character i.e. no 0005/NJ and no 0075/NK in code."""
    return signals and not check_0005(signals) and not check_0075(signals)



# -*- coding: utf-8 -*-
"""Style manager"""
from .global_settings import UI


class Styles(object):
    """Manage styles"""
    order = [x for x in 'rbislu']
    names = ['roman', 'bold', 'italic', 'smallcaps', 'inferior', 'superior']
    style_dict = dict(zip(order, names))

    def __init__(self, styles_string=None, allow_multiple=True,
                 manual_choice=False):
        self.styles_list = []
        self.allow_multiple = allow_multiple
        if manual_choice or styles_string is None:
            self.styles_string = styles_string or 'r'
            self.choose()
        elif styles_string:
            self.styles_string = styles_string
        else:
            # Setting an empty style will set all'
            self.styles_list = Styles.order

    def __iter__(self):
        return (x for x in self.styles_list)

    def __str__(self):
        if self.styles_list == Styles.order:
            return 'all'
        else:
            return ', '.join(self.values())

    def __repr__(self):
        return ''.join(self.styles_list)

    def __call__(self):
        return ''.join(self.styles_list)

    @property
    def styles_string(self):
        """Return a styles string"""
        return ''.join(self.styles_list)

    @styles_string.setter
    def styles_string(self, styles_string):
        """Set styles based on string"""
        styles_list = [x for x in Styles.order if x in styles_string]
        if self.allow_multiple:
            self.styles_list = styles_list
        else:
            self.styles_list = styles_list[:1]

    def keys(self):
        """Return a list of style letters"""
        return [x for x in self.styles_list]

    def values(self):
        """Return a list of style names for each style"""
        names = (Styles.style_dict.get(item, '') for item in self.styles_list)
        return [name for name in names if name]

    def items(self):
        """Returns a list of tuples: (char, name)"""
        return zip(self.keys(), self.values())

    def choose(self):
        """Chooses one or more styles and returns a list of them"""
        desc = ('\nAvailable options:\n' +
                '\n'.join('%s - %s' % (x, Styles.style_dict[x])
                          for x in Styles.order))
        header = 'Choose a text style.\n'
        if self.allow_multiple:
            header = ('Choose one or more text styles, '
                      'e.g. roman and small caps.\n')
        current = '\nCurrent: %s\n' % self
        UI.display(header + current + desc)
        self.styles_string = UI.enter_data_or_default('Your choice?',
                                                      self.styles_string)



# -*- coding: utf-8 -*-
"""
casting_session:

A module for everything related to working on a Monotype composition caster:
-casting composed type,
-punching paper tape (ribbon) for casters without interfaces,
-casting typecases based on character frequencies,
-casting a desired number of characters from matrix with x, y coordinates,
-composing and casting a line of text (not there yet)
-testing all valves, lines and pinblocks,
-calibrating the space transfer wedge, mould opening, diecase draw rods,
 position of character on type body
-sending any codes/combinations to the caster.
"""

# IMPORTS:
from collections import deque
from copy import copy
# Signals parsing methods for rpi2caster
from . import parsing as p
# Custom exceptions
from . import exceptions as e
# Constants shared between modules
from . import constants as c
# Typesetting functions module
from . import typesetting_funcs as tsf
# Style manager
from .styles import Styles
# Measure manager
from .measure import Measure
# Caster backend
from .monotype import MonotypeCaster
# Casting stats
from .casting_stats import Stats
# Globally selected UI
from .global_settings import UI
# Typecase casting needs this
from . import letter_frequencies
# Matrix, wedge and typesetting data models
from .typesetting_data import Ribbon
from .matrix_data import Diecase, diecase_operations
from .wedge_data import Wedge


def choose_sensor_and_driver(casting_routine):
    """Checks current modes (simulation, perforation, testing)"""
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        mode = self.caster.mode
        UI.debug_pause('About to %s...' %
                       (mode.casting and 'cast composition' or
                        mode.testing and 'test the outputs' or
                        mode.calibration and 'calibrate the machine' or
                        mode.punching and 'punch the ribbon'))
        # Instantiate and enter context
        with mode.sensor() as self.caster.sensor:
            with mode.output() as self.caster.output:
                with self.caster:
                    return casting_routine(self, *args, **kwargs)
    return wrapper


def cast_or_punch_result(ribbon_source):
    """Get the ribbon from decorated routine and cast it"""
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        mode = self.caster.mode
        try:
            ribbon = ribbon_source(self, *args, **kwargs)
            if ribbon:
                self.cast_codes(ribbon)
        except e.CastingAborted:
            pass
        finally:
            # Reset diagnostics and row 16 addressing modes
            for mode in (mode.kmn, mode.hmn, mode.unitshift,
                         mode.testing, mode.calibration):
                mode = False
    return wrapper


class Casting(object):
    """Casting:

    Methods related to operating the composition caster.
    Requires configured caster.

    All methods related to operating a composition caster are here:
    -casting composition and sorts, punching composition,
    -calibrating the caster,
    -testing the interface,
    -sending an arbitrary combination of signals,
    -casting spaces to heat up the mould."""

    def __init__(self, ribbon_file='', ribbon_id='', diecase_id='',
                 wedge_name=''):
        # Caster for this job
        self.caster = MonotypeCaster()
        self.stats = Stats(self)
        self.measure = Measure(manual=False)
        self.ribbon = Ribbon(filename=ribbon_file, ribbon_id=ribbon_id)
        if diecase_id:
            self.diecase = Diecase(diecase_id)
        if wedge_name:
            self.diecase.alt_wedge = Wedge(wedge_name)

    @choose_sensor_and_driver
    def cast_codes(self, ribbon):
        """
        Main casting routine.
        Cast or punch a sequence of codes, displaying the statistics.
        First choose a number of repetitions and lines skipped (for casting),
        diagnostics and punching will be processed from start to end.
        For each run, display statistics and send signals to the caster
        one by one, until a whole sequence is cast.
        If casting multiple runs, repeat until all are done.
        A run may not be successful (casting is aborted by machine stop
        or ctrl-C interrupt) - in this case, lines cast successfully
        may be skipped, and the rest will be cast.
        """
        mode = self.caster.mode
        # Rewind the ribbon if 0005 is found before 0005+0075
        if not (mode.testing or mode.punching) and p.stop_comes_first(ribbon):
            ribbon = [x for x in reversed(ribbon)]
        # New stats for the resulting ribbon
        self.stats.ribbon = ribbon
        UI.display_parameters({'Ribbon info': self.stats.ribbon_parameters})
        # Always 1 run for calibrating and punching
        if mode.diagnostics or mode.punching:
            self.stats.runs = 1
            l_skipped = 0
        else:
            prompt = 'How many times do you want to cast it?'
            self.stats.runs = abs(UI.enter_data_or_default(prompt, 1, int))
            # Line skipping - ask user if they want to skip any initial line(s)
            prompt = 'How many initial lines do you want to skip?'
            l_skipped = (self.stats.get_ribbon_lines() > 1 and
                         abs(UI.enter_data_or_default(prompt, 0, int)) or 0)
        UI.display_parameters({'Session info': self.stats.session_parameters})
        # For each casting run repeat
        while self.stats.get_runs_left():
            queue = deque(ribbon)
            # Apply constraints: 0 <= lines_skipped < lines in ribbon
            l_skipped = min(l_skipped, self.stats.get_ribbon_lines() - 1)
            l_skipped = max(0, l_skipped)
            if l_skipped:
                UI.display('Skipping %s lines' % l_skipped)
            # Take away combinations until we skip the desired number of lines
            # BEWARE: ribbon starts with galley trip!
            # We must give it back after lines are taken away
            code = ''
            while l_skipped + 1 > 0 and not mode.diagnostics:
                code = queue.popleft()
                l_skipped -= 1 * p.check_newline(code)
            queue.appendleft(code)
            # The ribbon is ready for casting / punching
            self.stats.queue = queue
            exit_prompt = '[Y] to start next run or [N] to exit?'
            if self.process_casting_run(queue):
                # Casting successful - ready to cast next run - ask to repeat
                # after the last run is completed (because user may want to
                # cast / punch once more?)
                if (self.stats.all_done() and
                        UI.confirm('One more run?', default=mode.diagnostics)):
                    self.stats.add_one_more_run()
            elif mode.casting and UI.confirm('Retry this run?', default=True):
                # Casting aborted - ask if user wants to repeat
                self.stats.undo_last_run()
                self.stats.add_one_more_run()
                lines_ok = self.stats.get_lines_done()
                prompt = 'Skip %s lines successfully cast?' % lines_ok
                if lines_ok > 0 and UI.confirm(prompt, default=True):
                    l_skipped = lines_ok
                # Start this run again
            elif (not self.stats.all_done() and
                  UI.confirm(exit_prompt, default=True)):
                # There are some more runs to do - go on?
                self.stats.undo_last_run()
            else:
                return

    def process_casting_run(self, casting_queue):
        """Casts the sequence of codes in given sequence"""
        mode = self.caster.mode
        generator = (p.parse_record(record) for record in casting_queue)
        if not mode.testing:
            self.caster.sensor.check_if_machine_is_working()
        while True:
            try:
                signals, comment = next(generator)
                if comment and not signals:
                    UI.display('\n\n' + comment + '\n' + '-' * len(comment))
                    continue
                self.stats.signals = signals
                UI.display_parameters({comment: self.stats.code_parameters})
                # Let the caster do the job
                self.caster.process(signals)
            except StopIteration:
                self.stats.end_run()
                return True
            except (e.MachineStopped, KeyboardInterrupt, EOFError):
                # Allow resume in punching mode
                if (mode.punching and not mode.diagnostics and
                        UI.confirm('Continue?', default=True)):
                    self.caster.process(signals)
                else:
                    self.stats.end_run()
                    return False

    @cast_or_punch_result
    def _test_front_pinblock(self):
        """Sends signals 1...14, one by one"""
        UI.pause('Testing the front pinblock - signals 1 towards 14.')
        self.caster.mode.testing = True
        return [str(n) for n in range(1, 15)]

    @cast_or_punch_result
    def _test_rear_pinblock(self):
        """Sends NI, NL, A...N"""
        UI.pause('This will test the front pinblock - signals NI, NL, A...N. ')
        self.caster.mode.testing = True
        return [x for x in c.COLUMNS_17]

    @cast_or_punch_result
    def _test_all(self):
        """Tests all valves and composition caster's inputs in original
        Monotype order: NMLKJIHGFSED 0075 CBA 123456789 10 11 12 13 14 0005.
        """
        UI.pause('This will test all the air lines in the same order '
                 'as the holes on the paper tower: \n%s\n'
                 'MAKE SURE THE PUMP IS DISENGAGED.' % ' '.join(c.SIGNALS))
        self.caster.mode.testing = True
        return [x for x in c.SIGNALS]

    @cast_or_punch_result
    def _test_justification(self):
        """Tests the 0075-S-0005"""
        UI.pause('This will test the justification pin block (0075, S, 0005).')
        self.caster.mode.testing = True
        return ['0075', 'S', '0005']

    @choose_sensor_and_driver
    def _test_any_code(self):
        """Tests a user-specified combination of signals"""
        self.caster.mode.testing = True
        while True:
            UI.display('Enter the signals to send to the caster, '
                       'or leave empty to return to menu: ')
            prompt = 'Signals? (leave blank to exit)'
            signals = p.parse_signals(UI.enter_data_or_blank(prompt))
            if signals:
                self.caster.output.valves_off()
                UI.display('Activating ' + ' '.join(signals))
                self.caster.output.valves_on(signals)
            else:
                break
        self.caster.mode.testing = False

    @cast_or_punch_result
    def cast_composition(self):
        """Casts or punches the ribbon contents if there are any"""
        if not self.ribbon.contents:
            e.return_to_menu()
        return self.ribbon.contents

    def cast_sorts(self):
        """Sorts casting routine, based on the position in diecase.
        Ask user about the diecase row & column, as well as number of sorts.
        """
        order = []
        while True:
            if self.diecase:
                char = UI.enter_data_or_blank('Character?')
                matrix = self.diecase.lookup_matrix(char)
                prompt = ('Width correction? (%.2f...%.2f points)'
                          % (matrix.get_min_points() - matrix.points,
                             matrix.get_max_points() - matrix.points))
                delta = UI.enter_data_or_default(prompt, 0, float)
                matrix.points += delta
            else:
                matrix = self.diecase.lookup_matrix()
                matrix.specify_units()
            qty = UI.enter_data_or_default('How many sorts?', 10, int)
            order.append((matrix, qty))
            prompt = 'More sorts? Otherwise, start casting'
            if not UI.confirm(prompt, default=True):
                break
        # Now let's calculate and cast it...
        self.cast_galley(order)

    def cast_typecases(self):
        """Casting typecases according to supplied font scheme."""
        enter = UI.enter_data_or_default
        freqs = letter_frequencies.CharFreqs()
        freqs.define_case_ratio()
        freqs.define_scale()
        supported_styles = self.diecase.styles
        UI.display('Styles to cast?')
        # Order specified numbers of sorts from desired matrices
        order = []
        style_manager = Styles(supported_styles)
        style_manager.choose()
        styles = style_manager.keys()
        for style, name in style_manager.items():
            # Display style name
            UI.display_header(name)
            if len(styles) == 1 or style == 'r':
                scale = 1.0
            else:
                scale = enter('Scale for %s?' % name, 100, float) / 100.0
            for char, chars_qty in freqs.type_bill:
                qty = int(scale * chars_qty)
                UI.display('%s: %s' % (char, chars_qty))
                matrix = self.diecase.lookup_matrix(char, style)
                order.append((matrix, qty))
        self.cast_galley(order)

    def cast_spaces(self):
        """Spaces casting routine, based on the position in diecase.
        Ask user about the space width and measurement unit.
        """
        order = []
        master = self.diecase.lookup_matrix(tsf.high_or_low_space())
        while True:
            matrix = copy(master)
            UI.display('Enter the space width...')
            UI.display('\nWidth limits for this matrix: %spt min - %spt max\n'
                       % (matrix.get_min_points(), matrix.get_max_points()))
            width = Measure(6, 'pt').points
            matrix.points = width
            prompt = 'How many lines?'
            lines = UI.enter_data_or_default(prompt, 1, int)
            order.extend([(matrix, 0)] * lines)
            prompt = 'More spaces? Otherwise, start casting'
            if not UI.confirm(prompt, default=True):
                break
        self.cast_galley(order)

    @cast_or_punch_result
    def cast_galley(self, order=()):
        """Cast a series of type, filling lines of given width to the brim.

        Each character is specified by a tuple: (matrix, qty)
            where matrix is a Matrix object,
            qty is quantity (0 for a filled line,
                             >0 for a given number of chars).

        If there is too many chars for a single line - will cast more lines.
        Last line will be quadded out.
        Characters other than low spaces will be separated by double G2 spaces
        to prevent matrices from overheating.
        """
        if not order:
            e.return_to_menu()
        # 1 quad before and after the line
        quad_padding = 1
        quad = self.diecase.decode_matrix('O15')
        # Leave some slack to adjust the line
        length = self.measure.points - 2 * quad_padding * quad.points
        # Build a sequence of matrices for casting
        # If n is 0, we fill the line to the brim
        queues = ([mat] * n if n else [mat] * int((length // mat.points) - 1)
                  for (mat, n) in order)
        matrix_stream = (mat for batch in queues for mat in batch)
        # Initialize the galley-constructor
        builder = tsf.GalleyBuilder(matrix_stream, self.diecase, self.measure)
        builder.quad_padding = quad_padding
        builder.cooldown = True
        builder.mould_heatup = UI.confirm('Pre-heat the mould?', True)
        job = self.caster.mode.punching and 'punching' or 'casting'
        UI.display('Generating a sequence for %s...' % job)
        queue = builder.build_galley()
        UI.display('\nReady for %s...\n\n' % job)
        UI.display('Each line will have two em-quads at the start '
                   'and at the end, to support the type.\n'
                   'Starting with two lines of quads to heat up the mould.\n')
        return queue

    @cast_or_punch_result
    def quick_typesetting(self):
        """Allows us to use caster for casting single lines.
        This means that the user enters a text to be cast,
        gives the line length, chooses alignment and diecase.
        Then, the program composes the line, justifies it, translates it
        to Monotype code combinations.

        This allows for quick typesetting of short texts, like names etc.
        """
        supported_styles = self.diecase.styles
        style = Styles(supported_styles, allow_multiple=False,
                       manual_choice=True)()
        text = UI.enter_data('Text to compose?')
        if not self.diecase.test_characters(text, style):
            UI.display('WARNING: Some characters are missing!')
            if not UI.confirm('Continue?', default=False):
                return
        space = self.diecase.decode_matrix('G2')
        matrix_stream = (self.diecase.lookup_matrix(char, style) if char != ' '
                         else space for char in text)
        builder = tsf.GalleyBuilder(matrix_stream, self.diecase, self.measure)
        builder.mould_heatup = False
        queue = builder.build_galley()
        UI.display('Each line will have two em-quads at the start '
                   'and at the end, to support the type.\n'
                   'Starting with two lines of quads to heat up the mould.\n')
        return queue

    @cast_or_punch_result
    def _calibrate_wedges(self):
        """Allows to calibrate the justification wedges so that when you're
        casting a 9-unit character with the S-needle at 0075:3 and 0005:8
        (neutral position), the    width is the same.

        It works like this:
        1. 0075 - turn the pump on,
        2. cast 10 spaces from the specified matrix (default: G9),
        3. put the line to the galley & set 0005 to 8, 0075 to 3, pump on,
        4. cast 10 spaces with the S-needle from the same matrix,
        5. put the line to the galley, then 0005 to turn the pump off.
        """
        UI.display('Transfer wedge calibration:\n\n'
                   'This function will cast two lines of 5 spaces: '
                   'first: G5, second: GS5 with wedges at 3/8. \n'
                   'Adjust the 52D space transfer wedge '
                   'until the lengths are the same.')
        if not UI.confirm('\nProceed?', default=True):
            return None
        mat = self.diecase.decode_matrix('G5')
        self.caster.mode.calibration = True
        sequence = tsf.end_casting()
        sequence.extend(['S %s with S-needle' % mat] * 7)
        sequence.extend(['%s' % mat] * 7)
        sequence.extend(tsf.double_justification())
        return sequence

    @cast_or_punch_result
    def _calibrate_mould(self):
        """Calculates the width, displays it and casts some 9-unit characters.
        Then, the user measures the width and adjusts the mould opening width.
        """
        wedge = self.diecase.alt_wedge
        UI.display('Mould blade opening calibration:\n'
                   'Cast G5 (9-units wide on S5 wedge), then measure '
                   'the width. Adjust if needed.')
        UI.display_parameters({'Wedge data': wedge.parameters})
        UI.display('\n')
        UI.display('9 units (1en) is %s" wide' % round(wedge.em_width / 2, 4))
        UI.display('18 units (1em) is %s" wide\n' % round(wedge.em_width, 4))
        if not UI.confirm('\nProceed?', default=True):
            return None
        mat = self.diecase.decode_matrix('G5')
        self.caster.mode.calibration = True
        sequence = tsf.end_casting() + ['%s' % mat] * 7
        sequence.extend(tsf.double_justification())
        return sequence

    @cast_or_punch_result
    def _calibrate_diecase(self):
        """Casts the "en dash" characters for calibrating the character X-Y
        relative to type body."""
        UI.display('X-Y character calibration:\n'
                   'Cast some en-dashes and/or lowercase "n" letters, '
                   'then check the position of the character relative to the '
                   'type body.\nAdjust if needed.')
        self.caster.mode.calibration = True
        # Build character list
        queue = tsf.end_casting()
        wedge_positions = (3, 8)
        for char in ('--', 'n', 'h'):
            mat = self.diecase.lookup_matrix(char)
            if not self.diecase:
                mat.specify_units()
            # Change justification wedge positions if they are different
            # (only if the type was cast with corrections)
            prev_wedge_positions = wedge_positions
            wedge_positions = mat.wedge_positions()
            wedges_change = wedge_positions != prev_wedge_positions
            if prev_wedge_positions != (3, 8) and wedges_change:
                queue.extend(tsf.single_justification(prev_wedge_positions))
            # Add the characters to the queue
            queue.extend(['%s' % mat] * 3)
        # Set the initial wedge positions
        queue.extend(tsf.double_justification(wedge_positions))
        return queue

    @choose_sensor_and_driver
    def _calibrate_draw_rods(self):
        """Keeps the diecase at G8 so that the operator can adjust
        the diecase draw rods until the diecase stops moving sideways
        when the centering pin is descending."""
        UI.display('Draw rods calibration:\n'
                   'The diecase will be moved to the central position (G-8), '
                   'turn on the machine\nand adjust the diecase draw rods '
                   'until the diecase stops moving sideways as the\n'
                   'centering pin is descending into the hole in the matrix.')
        if not UI.confirm('\nProceed?', default=True):
            return None
        self.caster.output.valves_on(['G', '8'])
        UI.pause('Sending G8, waiting for you to stop...')

    @cast_or_punch_result
    def _test_row_16(self):
        """Tests the 16th row with selected addressing mode
        (HMN, KMN, unit-shift). Casts from all matrices in 16th row."""
        UI.display('This will test the 16th row addressing.\n'
                   'If your caster has HMN, KMN or unit-shift attachment, '
                   'turn it on.\n')
        if not UI.confirm('\nProceed?', default=True):
            return None
        self.caster.mode.calibration = True
        queue = tsf.end_casting() + ['%s16' % x for x in c.COLUMNS_17]
        queue.extend(tsf.double_justification(comment='Starting...'))
        return queue

    def diagnostics_submenu(self):
        """Settings and alignment menu for servicing the caster"""
        def finish():
            """Sets the flag to True"""
            nonlocal finished
            finished = True

        def menu_options():
            """Build a list of options, adding an option if condition is met"""
            caster = not self.caster.mode.punching
            opts = [(finish, 'Back', 'Returns to main menu', True),
                    (self._test_all, 'Test outputs',
                     'Test all the air outputs N...O15, one by one', True),
                    (self._test_front_pinblock, 'Test the front pin block',
                     'Test the pins 1...14', caster),
                    (self._test_rear_pinblock, 'Test the rear pin block',
                     'Test the pins NI, NL, A...N, one by one', caster),
                    (self._test_justification, 'Test the justification block',
                     'Test the pins for 0075, S and 0005, one by one', caster),
                    (self._test_any_code, 'Send specified signal combination',
                     'Send the specified signals to the machine', True),
                    (self._calibrate_wedges, 'Calibrate the 52D wedge',
                     'Calibrate the space transfer wedge for correct width',
                     caster),
                    (self._calibrate_mould, 'Calibrate mould opening',
                     'Cast 9-unit characters to adjust the type width',
                     caster),
                    (self._calibrate_draw_rods, 'Calibrate diecase draw rods',
                     'Keep the matrix case at G8 and adjust the draw rods',
                     caster),
                    (self._calibrate_diecase, 'Calibrate matrix X-Y',
                     'Calibrate the character-to-body positioning', caster),
                    (self._test_row_16, 'Test HMN, KMN or unit-shift',
                     'Cast type from row 16 with chosen addressing mode',
                     caster)]
            return [(function, description, long_description) for
                    (function, description, long_description, condition)
                    in opts if condition]

        header = ('Diagnostics and machine calibration menu:\n\n')
        # Keep displaying the menu and go back here after any method ends
        finished = False
        while not finished:
            try:
                # Catch "return to menu" and "exit program" exceptions here
                UI.menu(menu_options(), header=header)()
            except e.ReturnToMenu:
                # Stay in the menu
                pass

    def main_menu(self):
        """Main menu for the type casting utility."""
        def finish():
            """Sets the flag to True"""
            nonlocal finished
            finished = True

        def choose_ribbon():
            """Chooses a ribbon from database or file"""
            self.ribbon = Ribbon(manual_choice=True)

        def choose_diecase():
            """Chooses a diecase from database"""
            self.diecase = Diecase(manual_choice=True)

        def choose_wedge():
            """Chooses a wedge from registered ones"""
            self.diecase.alt_wedge = Wedge(manual_choice=True)

        def choose_measure():
            """Chooses a galley width"""
            self.measure = Measure()

        def menu_options():
            """Build a list of options, adding an option if condition is met"""
            # Options are described with tuples:
            # (function, description, condition)
            caster = not self.caster.mode.punching
            diecase = bool(self.diecase)
            ribbon = bool(self.ribbon)
            diecase_info = diecase and ' (current: %s)' % diecase or ''
            opts = [(finish, 'Exit', 'Exits the program', True),
                    (self.cast_composition, 'Cast composition',
                     'Cast type from a selected ribbon', ribbon and caster),
                    (self.cast_composition, 'Punch ribbon',
                     'Punch a paper ribbon for casting without the interface',
                     ribbon and not caster),
                    (choose_ribbon, 'Select ribbon',
                     'Select a ribbon from database or file', True),
                    (choose_diecase, 'Select diecase',
                     'Select a matrix case from database' + diecase_info,
                     caster),
                    (choose_wedge, 'Select wedge',
                     'Enter a wedge designation (current: %s)'
                     % self.diecase.alt_wedge, caster),
                    (choose_measure, 'Change galley width',
                     'Set a new galley width for casting type (current: %s)'
                     % self.measure, caster),
                    (self.ribbon.display_contents, 'View codes',
                     'Display all codes in the selected ribbon', ribbon),
                    (self.diecase.show_layout, 'Show diecase layout',
                     'View the matrix case layout', diecase and caster),
                    (self.quick_typesetting, 'Quick typesetting',
                     'Compose and cast a line of text', self.diecase),
                    (self.cast_sorts, 'Cast sorts for given characters',
                     'Cast from matrix based on a character',
                     caster and diecase),
                    (self.cast_sorts, 'Cast sorts from matrix coordinates',
                     'Cast from matrix at given position',
                     caster and not diecase),
                    (self.cast_spaces, 'Cast spaces or quads',
                     'Cast spaces or quads of a specified width', caster),
                    (self.cast_typecases, 'Cast typecases',
                     'Cast a typecase based on a selected language',
                     caster and diecase),
                    (self._display_details, 'Show details...',
                     'Display ribbon, diecase, wedge and interface info',
                     caster),
                    (self._display_details, 'Show details...',
                     'Display ribbon and interface details', not caster),
                    (diecase_operations, 'Matrix manipulation...',
                     'Work on matrix cases', True),
                    (self.diagnostics_submenu, 'Service...',
                     'Interface and machine diagnostic functions', True)]
            # Built a list of menu options conditionally
            return [(function, description, long_description)
                    for (function, description, long_description, condition)
                    in opts if condition]

        header = ('rpi2caster - CAT (Computer-Aided Typecasting) '
                  'for Monotype Composition or Type and Rule casters.\n\n'
                  'This program reads a ribbon (from file or database) '
                  'and casts the type on a composition caster.'
                  '\n\nCasting / Punching Menu:')
        # Keep displaying the menu and go back here after any method ends
        finished = False
        while not finished:
            # Catch any known exceptions here
            try:
                UI.menu(menu_options(), header=header, footer='')()
            except (e.ReturnToMenu, e.MenuLevelUp, KeyboardInterrupt):
                # Will skip to the end of the loop, and start all over
                pass

    @property
    def ribbon(self):
        """Ribbon for the casting session"""
        return self.__dict__.get('_ribbon') or Ribbon()

    @ribbon.setter
    def ribbon(self, ribbon):
        """Ribbon setter"""
        self.__dict__['_ribbon'] = ribbon or Ribbon()
        if ribbon.diecase_id:
            self.diecase = Diecase(ribbon.diecase_id)
        if ribbon.wedge_name:
            self.diecase.alt_wedge = Wedge(ribbon.wedge_name)

    @property
    def diecase(self):
        """Diecase for the casting session"""
        return self.__dict__.get('_diecase') or Diecase()

    @diecase.setter
    def diecase(self, diecase):
        """Diecase setter"""
        self.__dict__['_diecase'] = diecase

    def _display_details(self):
        """Collect ribbon, diecase and wedge data here"""
        display = UI.display_parameters
        punching = self.caster.mode.punching
        if self.ribbon:
            display({'Ribbon data': self.ribbon.parameters})
        if self.diecase and not punching:
            display({'Matrix case data': self.diecase.parameters})
        if self.diecase.alt_wedge and not punching:
            display({'Wedge data': self.diecase.alt_wedge.parameters})
        display({'Caster data': self.caster.parameters})
        UI.pause()



# -*- coding: utf-8 -*-
"""Casting stats - implements stats getting and setting"""
from . import parsing as p


class Stats(object):
    """Casting statistics gathering and displaying functions"""
    def __init__(self, session):
        self.__dict__['_current'] = {}
        self.__dict__['_previous'] = {}
        self.__dict__['_run'] = {}
        self.__dict__['_ribbon'] = {}
        self.__dict__['_session'] = {}
        self.session = session

    def end_run(self):
        """Updates the info about the runs"""
        session = self.__dict__['_session']
        # Current run will always be runs done +1
        session['runs_done'] = session.get('current_run', 1)
        session['current_run'] = session.get('current_run', 1) + 1

    def all_done(self):
        """Check if there are runs left"""
        session = self.__dict__['_session']
        return session.get('runs', 1) - session.get('runs_done', 0) <= 0

    def add_one_more_run(self):
        """Adds one to all_runs number - in case of repeating the job"""
        # Updating the attribute to trigger the "magic" in self.runs setter
        self.runs += 1

    def undo_last_run(self):
        """Subtracts the current run stats from all runs stats;
        this is called before repeating a failed casting run."""
        session = self.__dict__['_session']
        run = self.__dict__['_run']
        session['failed_runs'] = session.get('failed_runs', 0) + 1
        for par in ('current_line', 'current_code', 'current_char'):
            session[par] = (session.get(par, 1) - run.get(par, 0))

    def get_runs_left(self):
        """Gets the runs left number"""
        session = self.__dict__['_session']
        return session.get('runs', 1) - session.get('runs_done', 0)

    def get_lines_done(self):
        """Gets the current run line number"""
        return self.__dict__.get('_run', {}).get('current_line', 0) - 1

    def get_ribbon_lines(self):
        """Gets number of lines per ribbon"""
        return self.__dict__.get('_ribbon', {}).get('lines', 1)

    def get_current_run(self):
        """Gets the current run number"""
        return self.__dict__.get('_session', {}).get('current_run', 1)

    @property
    def runs(self):
        """Gets the runs (repetitions) number"""
        return self.__dict__.get('_session', {}).get('runs', 1)

    @runs.setter
    def runs(self, runs=1):
        """Sets the runs (repetitions) number"""
        session = self.__dict__['_session']
        ribbon = self.__dict__['_ribbon']
        # Update whenever we change runs number via one_more_run
        session['runs'] = runs
        session['codes'] = runs * ribbon.get('codes', 1)
        session['chars'] = runs * ribbon.get('chars', 1)
        session['lines'] = runs * ribbon.get('lines', 1)

    @property
    def ribbon_parameters(self):
        """Gets ribbon parameters"""
        ribbon = self.__dict__.get('_ribbon', {})
        if self.session.caster.mode.diagnostics:
            return []
        else:
            return [(ribbon.get('codes', 1), 'Combinations in ribbon'),
                    (ribbon.get('lines', 1), 'Lines to cast'),
                    (ribbon.get('chars', 1), 'Characters incl. spaces')]

    @property
    def session_parameters(self):
        """Displays the ribbon data"""
        session = self.__dict__.get('_session', {})
        # Don't display in single-run sessions or calibration mode
        if self.session.caster.mode.diagnostics or session.get('runs', 1) < 2:
            return []
        else:
            return [(session.get('runs', 1), 'Casting runs'),
                    (session.get('codes', 1), 'All codes'),
                    (session.get('chars', 1), 'All chars'),
                    (session.get('lines', 1), 'All lines')]

    @property
    def code_parameters(self):
        """Displays info about the current combination"""
        # Aliases to keep code a bit cleaner:
        current = self.__dict__.get('_current', {})
        run = self.__dict__.get('_run', {})
        session = self.__dict__.get('_session', {})
        # Some options are viable only if we have more than one repetition
        multiple_runs = session.get('runs', 1) > 1
        # What to do with current code: is it a char or newline?
        is_new_line = p.check_newline(current.get('signals', []))
        # Process the data to display, starting with general data
        data = [(' '.join(current.get('signals', [])), 'Signals from ribbon')]
        # For casting and punching:
        if not self.session.caster.mode.diagnostics:
            # Codes per run
            if multiple_runs:
                data.append(build_data(run, 'code', 'Code / run'))
            # Codes per session
            data.append(build_data(session, 'code', 'Code / job'))
            # Runs per session
            if multiple_runs:
                data.append(build_data(session, 'run', 'Run / job'))
            data.append((session.get('failed_runs', 0), 'Failed runs'))
        if self.session.caster.mode.casting:
            # Displayed pump status
            pump_bool = current.get('pump_working', False)
            pump = {True: 'ON', False: 'OFF'}[pump_bool]
            data.append((is_new_line and run.get('current_line', 1),
                         'Starting a new line'))
            # Characters per run
            if multiple_runs:
                data.append(build_data(run, 'char', 'Character / run'))
            # Characters per session
            data.append(build_data(session, 'char', 'Character / job'))
            # Lines per session
            if multiple_runs:
                data.append(build_data(run, 'line', 'Line / run'))
            # Lines per run
            data.append(build_data(session, 'line', 'Line / job'))
            data.append((current.get('0075', '15'), 'Wedge 0075 now at'))
            data.append((current.get('0005', '15'), 'Wedge 0005 now at'))
            data.append((pump, 'Pump is'))
        return data

    @session_parameters.setter
    def ribbon(self, ribbon_contents):
        """Parses the ribbon, counts combinations, lines and characters"""
        mode = self.session.caster.mode
        # Reset counters
        self.__dict__['_ribbon'] = {'lines': -1, 'codes': 0, 'chars': 0}
        self.__dict__['_session'] = {'lines': 0, 'codes': 0, 'chars': 0}
        self.__dict__['_current'] = {}
        self.__dict__['_previous'] = {}
        ribbon = self.__dict__['_ribbon']
        code_generator = (code for (code, _) in
                          (p.parse_record(x) for x in ribbon_contents) if code)
        for code in code_generator:
            # Guards against empty combination i.e. line with comment only
            ribbon['codes'] += 1
            # When row 16 is encountered, ask user for addressing mode
            # (depends on the machine - most common is 15x17 and unit-shift)
            row_16_supported = mode.unitshift or mode.hmn or mode.kmn
            if '16' in code and not row_16_supported:
                # Do it now - before casting starts!
                mode.choose_row16_addressing()
            if p.check_newline(code):
                ribbon['lines'] += 1
            elif p.check_character(code):
                ribbon['chars'] += 1

    @ribbon_parameters.setter
    def queue(self, queue):
        """Parses the ribbon, counts combinations, lines and characters"""
        # Clear and start with -1 line for the initial galley trip
        self.__dict__['_run'] = {'lines': -1, 'codes': 0, 'chars': 0}
        run = self.__dict__['_run']
        # Before first casting, set the current run number to 1
        # (the session current run will be unset yet, so default)
        session = self.__dict__['_session']
        session['current_run'] = session.get('current_run', 1)
        code_generator = (code for (code, _) in
                          (p.parse_record(x) for x in queue) if code)
        for code in code_generator:
            # Guards against empty combination i.e. line with comment only
            run['codes'] += 1
            if p.check_newline(code):
                run['lines'] += 1
            elif p.check_character(code):
                run['chars'] += 1

    @code_parameters.setter
    def signals(self, signals):
        """Updates the stats based on current combination"""
        if not signals:
            return False
        # Save previous state
        current = self.__dict__['_current']
        run = self.__dict__['_run']
        session = self.__dict__['_session']
        self.__dict__['_previous'] = {k: v for k, v in current.items()}
        # Update combination info
        current['signals'] = signals
        run['current_code'] = run.get('current_code', 0) + 1
        session['current_code'] = session.get('current_code', 0) + 1
        if p.check_character(signals):
            run['current_char'] = run.get('current_char', 0) + 1
            session['current_char'] = session.get('current_char', 0) + 1
        elif self._check_double_justification():
            run['current_line'] = run.get('current_line', 0) + 1
            session['current_line'] = session.get('current_line', 0) + 1
        # Check the pump working/non-working status in the casting mode
        if self.session.caster.mode.casting:
            self._update_wedge_positions()
            self._check_pump()
        return True

    def _check_pump(self):
        """Checks pump based on current and previous combination"""
        prev_signals = self.__dict__.get('_previous', {}).get('signals', [])
        curr_signals = self.__dict__.get('_current', {}).get('signals', [])
        # Was it running until now? Get it from the caster
        running = self.session.caster.pump_working
        # Was it started before (0075 with or without 0005)?
        started = p.check_0075(prev_signals)
        # Was it stopped (0005 without 0075)
        stopped = p.check_0005(prev_signals) and not p.check_0075(prev_signals)
        # Is 0005 or 0075 in current combination? (if so - temporary stop)
        on_hold = p.check_0005(curr_signals) or p.check_0075(curr_signals)
        # Determine the current status
        pump_status = (running or started) and not stopped and not on_hold
        current = self.__dict__.get('_current', {})
        current['pump_working'] = pump_status
        # Feed it back to the caster object
        self.session.caster.pump_working = current.get('pump_working', False)

    def _update_wedge_positions(self):
        """Gets current positions of 0005 and 0075 wedges"""
        combination = self.__dict__.get('_current', {}).get('signals', [])
        # Check 0005
        if p.check_0005(combination):
            candidates = [x for x in range(15) if str(x) in combination]
            self.__dict__['_current']['0005'] = (candidates and
                                                 str(max(candidates)) or '15')
        # Check 0075
        if p.check_0075(combination):
            candidates = [x for x in range(15) if str(x) in combination]
            self.__dict__['_current']['0075'] = (candidates and
                                                 str(max(candidates)) or '15')

    def _check_double_justification(self):
        """Checks the current and previous signals for 0075+0005 -> 0075
        transition (which is used for detecting new line accurately)."""
        before = self.__dict__.get('_previous', {}).get('signals', [])
        now = self.__dict__.get('_current', {}).get('signals', [])
        return p.check_newline(before) and p.check_pump_start(now)


def build_data(source, parameter, data_name=''):
    """Builds data to display based on the given parameter"""
    try:
        percentage = ((source.get('current_%s' % parameter, 1) - 1) /
                      source.get('%ss' % parameter, 1) * 100)
    except ZeroDivisionError:
        percentage = 0
    return (('%s / %s [%.1f%% done], %s left'
             % (source.get('current_%s' % parameter, 0),
                source.get('%ss' % parameter, 0), percentage,
                source.get('%ss' % parameter, 1) -
                source.get('current_%s' % parameter, 1) + 1),
             data_name or parameter.capitalize()))



# -*- coding: utf-8 -*-
"""Caster object for either real or virtual Monotype composition caster"""
# Built-in time module
from time import time, sleep
# Custom exceptions module
from . import exceptions as e
# Constants module
from . import constants as c
# Default user interface
from .global_settings import UI
# Configuration parsing
from . import cfg_parser
# Constants for readability
AIR_ON = True
AIR_OFF = False
# Get the signals arrangement
try:
    SIGNALS = (str(cfg_parser.get_config('SignalsArrangements',
                                         'signals_arrangement')).upper())
except e.NotConfigured:
    SIGNALS = c.ALNUM_ARR
# Get the interface to use
try:
    SENSOR = str(cfg_parser.get_config('Control', 'sensor')).lower()
    OUTPUT = str(cfg_parser.get_config('Control', 'output')).lower()
except e.NotConfigured:
    SENSOR = 'simulation'
    OUTPUT = 'simulation'


def adjust_signals(worker_function):
    """Adjusts the signals received by caster so that:
    1. O15 is stripped unless we are in testing mode (if explicitly stated)
        or in punching mode (always - it helps drive the punches)
    2. row 16 is addressed according to the selected method: HMN, unit-shift
        or is not addressed for plain machines (O15 will be cast instead)
    """
    def wrapper(self, signals, *args, **kw):
        """Changes the signals to match the selected method of addressing
        16th row on the matrix case"""

        def convert_unitshift(signals):
            """Adjust signals for unit-shift:
            replace D column with EF;
            add D for shifting the diecase to row 16.
            Normal wedge stays where it was; only the diecase moves.
            The width of characters in row 16 is the same as for row 15,
            unless 0005 & 0075 wedge corrections are applied.
            """
            if 'D' in signals:
                signals.remove('D')
                signals.extend(['E', 'F'])
            if '16' in signals:
                signals.append('D')

        def convert_hmn(signals):
            """Adjust signals for HMN as commented below.
            The HMN attachment was rare and required special normal wedges
            with additional step for 16th row."""
            if 'H' in signals:
                # H -> HN
                signals.append('N')
            elif 'N' in signals and not ('I' in signals or 'L' in signals):
                # N -> MN
                signals.append('M')
            elif 'N' in signals or 'M' in signals:
                # NI, NL, M -> HNI, HNL, HM
                signals.append('H')
            elif [x for x in 'ABCDEFGIJKL' if x in signals]:
                # A...G, I...L -> HMA...G, HMI...L
                signals.extend(['H', 'M'])
            else:
                # O15 (no other signals) -> HMN
                signals.extend(['H', 'M', 'N'])

        def convert_kmn(signals):
            """Adjust signals for the rare KMN attachment.
            This attachment was made by one of Monotype's customers,
            and HMN was based off it - so it's even less common..."""
            if 'K' in signals:
                # K16: add N
                signals.append('N')
            elif 'M' in signals:
                # M16: add K
                signals.append('K')
            elif 'N' in signals and ('I' in signals or 'L' in signals):
                # NI16, NL16: add K
                signals.append('K')
            elif 'N' in signals:
                # N16: add M
                signals.append('M')
            elif [x for x in 'ABCDEFGHIJL' if x in signals]:
                # All signals apart from above and O16: add KM
                signals.extend(['K', 'M'])
            else:
                # O16: add KMN
                signals.extend(['K', 'M', 'N'])

        # No signals = no casting!
        if not signals:
            return False
        if self.mode.hmn and '16' in signals:
            convert_hmn(signals)
        elif self.mode.kmn and '16' in signals:
            convert_kmn(signals)
        elif self.mode.unitshift:
            # Unit-shift is active for all rows, not just 16
            convert_unitshift(signals)
        # Return a list of "normal" signals i.e. with no signal 16,
        # arranged in Monotype order (NMLK...13, 14, 0005, O15)
        # Scrap O15 signal unless we're in testing mode, where it can
        # be explicitly turned on.
        signals = [sig for sig in c.SIGNALS if sig in signals and
                   (sig is not 'O15' or self.mode.testing)]
        # Always add O15 for punching ribbons
        if self.mode.punching and 'O15' not in signals:
            signals.append('O15')
        # Adjust the signals just before casting; show the new combination
        UI.display(signals and ('Sending: ' + ' '.join(signals)) or '')
        return worker_function(self, signals, *args, **kw)
    return wrapper


class MonotypeCaster(object):
    """Methods common for Caster classes, used for instantiating
    caster driver objects (whether real hardware or mockup for simulation)."""
    def __init__(self):
        self.mode = CasterMode()
        self.sensor = SimulationSensor()
        self.output = SimulationOutput()
        self.lock = False
        self.pump_working = False

    def __enter__(self):
        """Lock the resource so that only one object can use it
        with context manager"""
        if self.lock:
            UI.display('Caster is already busy!')
        else:
            self.lock = True
            UI.debug_pause('Entering the caster context...')
            return self

    def __exit__(self, *_):
        UI.debug_pause('Caster no longer in use.')
        self.lock = False

    @property
    def parameters(self):
        """Gets a list of parameters"""
        # Collect data from I/O drivers
        return self.sensor.parameters + self.output.parameters

    @adjust_signals
    def process(self, signals):
        """process(signals):

        Checks for the machine's rotation, sends the signals (activates
        solenoid valves) after the caster is in the "air bar down" position.

        If no machine rotation is detected (sensor input doesn't change
        its state) during cycle_timeout, calls a function to ask user
        what to do (can be useful for resuming casting after manually
        stopping the machine for a short time - not recommended as the
        mould cools down and type quality can deteriorate).

        If the pump was working, the program will store 0075/0005 wedge
        positions, send 0005 to stop the pump and display the menu.
        If the operator decides to go on with casting, the aborted sequence
        will be re-cast so as to avoid missing characters in the composition.
        If the pump was working before, its operation will be resumed
        and the wedges will be reset to previous positions.

        When casting, the pace is dictated by the caster and its RPM. Thus,
        we can't arbitrarily set the intervals between valve ON and OFF
        signals. We need to get feedback from the machine, and we can use
        contact switch (unreliable in the long run), magnet & reed switch
        (not precise enough) or a photocell sensor + LED (very precise).
        We can use a paper tower's operating lever for obscuring the sensor
        (like John Cornelisse did), or we can use a partially obscured disc
        attached to the caster's shaft (like Bill Welliver did).
        Both ways are comparable; the former can be integrated with the
        valve block assembly, and the latter allows for very precise tweaking
        of duty cycle (bright/dark area ratio) and phase shift (disc's position
        relative to 0 degrees caster position).
        """
        busy = True
        while busy:
            # Escape this only by returning True on success,
            # or raising exceptions.CastingAborted, exceptions.ExitProgram
            # (which will be handled by the methods of the Casting class)
            try:
                # Casting cycle
                # (sensor on - valves on - sensor off - valves off)
                self.sensor.wait_for(AIR_ON)
                self.output.valves_on(signals)
                self.sensor.wait_for(AIR_OFF)
                self.output.valves_off()
                busy = False
                # Successful ending - the combination has been cast
            except (e.MachineStopped, KeyboardInterrupt, EOFError):
                # Machine stopped during casting - clean up
                self.pump_stop()
                # Exception will be handled in session
                raise e.MachineStopped

    def pump_stop(self):
        """Forces pump stop - won't end until it is turned off"""
        UI.display('Turning all valves off - just in case...')
        self.output.valves_off()
        while self.pump_working:
            UI.display('The pump is still working - turning it off...')
            try:
                # Run a full machine cycle to turn the pump off
                self.sensor.wait_for(AIR_ON, 30)
                self.output.valves_on(['N', 'J', 'S', '0005'])
                self.sensor.wait_for(AIR_OFF, 30)
                self.output.valves_off()
                UI.display('Pump is now off.')
                return True
            except (e.MachineStopped, KeyboardInterrupt, EOFError):
                pass


class SimulationSensor(object):
    """Mockup for a machine cycle sensor"""
    def __init__(self):
        self.lock = False
        self.manual_mode = True
        self.last_state = False
        self.name = 'mockup machine cycle sensor'

    def __enter__(self):
        if not self.lock:
            self.lock = True
            UI.debug_pause('Using a %s for machine feedback' % self.name)
            return self

    def __exit__(self, *_):
        UI.debug_pause('The %s is no longer in use' % self.name)
        # Reset manual mode
        self.manual_mode = True
        self.lock = False

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self.name, 'Sensor driver'),
                (self.manual_mode, 'Manual mode')]

    def check_if_machine_is_working(self):
        """Detect machine cycles and alert if it's not working"""
        UI.display('Now checking if the machine is running...')
        cycles = 3
        while True:
            try:
                while cycles:
                    # Run a new cycle
                    UI.display(cycles)
                    self.wait_for(new_state=True, timeout=30)
                    cycles -= 1
                return True
            except (e.MachineStopped, KeyboardInterrupt, EOFError):
                if not UI.confirm('Machine not running - do you want to '
                                  'continue? ', default=False):
                    e.return_to_menu()

    def wait_for(self, new_state, timeout=30, time_on=0.1, time_off=0.1):
        """Waits for a keypress to emulate machine cycle, unless user
        switches to auto mode, where all combinations are processed in batch"""
        status = {True: 'ON', False: 'OFF'}
        UI.debug_info('The sensor is going %s' % status[new_state])
        if self.manual_mode:
            start_time = time()
            # Ask whether to cast or simulate machine stop
            prompt = ('[A] to switch to automatic mode, [S] to stop\n'
                      'or leave blank to continue?')
            answer = UI.enter_data_or_blank(prompt) or ' '
            if answer in 'aA':
                self.manual_mode = False
            elif answer in 'sS':
                raise e.MachineStopped
            elif time() - start_time > timeout:
                UI.display('Timeout - you answered after %ds' % timeout)
                raise e.MachineStopped
        elif new_state:
            sleep(time_off)
        else:
            sleep(time_on)


class PunchingSensor(SimulationSensor):
    """A special sensor class for perforators"""
    def __init__(self):
        super().__init__()
        self.name = 'Timer-driven or manual advance for perforator'

    def check_if_machine_is_working(self):
        """Ask for user confirmation before punching"""
        UI.pause('\nRibbon punching: \n'
                 'Put the ribbon on the perforator and turn on the air.')

    def wait_for(self, new_state, timeout=30, time_on=0.25, time_off=0.4):
        """Calls simulation sensor's function, but with different timings"""
        super().wait_for(new_state, timeout, time_on, time_off)


class TestSensor(SimulationSensor):
    """A keyboard-operated "sensor" for testing inputs.
    No automatic mode is supported."""
    def __init__(self):
        super().__init__()
        self.name = 'Timer-driven or manual advance for perforator'

    def wait_for(self, new_state, *_):
        """Waits for keypress before turning the line off"""
        if not new_state:
            UI.pause('Next combination?')


class SimulationOutput(object):
    """Mockup for a driver for 32 pneumatic outputs"""
    def __init__(self, sig_arr=SIGNALS):
        self.lock = False
        self.name = 'mockup output driver for simulation'
        self.signals_arrangement = [str(x).upper() for x in sig_arr.split(',')]

    def __del__(self):
        UI.debug_pause('Deleting the %s' % self.name)

    def __enter__(self):
        if not self.lock:
            self.lock = True
            UI.debug_pause('Using the %s for sending signals...' % self.name)
            return self

    def __exit__(self, *_):
        self.valves_off()
        UI.debug_pause('Driver for %s no longer in use.' % self.name)
        self.lock = False

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self.name, 'Output driver'),
                (' '.join(self.signals_arrangement), 'Signals arrangement')]

    def one_on(self, sig):
        """Looks a signal up in arrangement and turns it on"""
        try:
            UI.debug_info(sig + ' on')
        except KeyError:
            raise e.WrongConfiguration('Signal %s not defined!' % sig)

    def one_off(self, sig):
        """Looks a signal up in arrangement and turns it on"""
        try:
            UI.debug_info(sig + ' off')
        except KeyError:
            raise e.WrongConfiguration('Signal %s not defined!' % sig)

    def valves_on(self, signals_list):
        """Turns on multiple valves"""
        for sig in signals_list:
            self.one_on(sig)

    def valves_off(self):
        """Turns off all the valves"""
        for sig in self.signals_arrangement:
            self.one_off(sig)


class CasterMode(object):
    """Session mode: casting / simulation / perforation"""
    def __init__(self):
        self.simulation = False
        self.punching = False
        self.testing = False
        self.calibration = False

    @property
    def simulation(self):
        """Simulation mode"""
        return self.__dict__.get('simulation', False)

    @simulation.setter
    def simulation(self, value):
        """Set the simulation mode"""
        self.__dict__['simulation'] = value and True or False

    @property
    def calibration(self):
        """Machine calibration mode"""
        return self.__dict__.get('calibration', False)

    @calibration.setter
    def calibration(self, value):
        """Set the machine calibration mode"""
        self.__dict__['calibration'] = value and True or False

    @property
    def diagnostics(self):
        """Machine diagnostics i.e. testing or calibration"""
        return self.testing or self.calibration

    @diagnostics.setter
    def diagnostics(self, value=False):
        """Turns the testing or calibration mode off"""
        if not value:
            self.testing = self.calibration = False

    @property
    def punching(self):
        """Punching mode"""
        return self.__dict__.get('punching', False)

    @punching.setter
    def punching(self, value):
        """Set the punching mode"""
        self.__dict__['punching'] = value and True or False

    @property
    def testing(self):
        """Testing mode"""
        return self.__dict__.get('testing', False)

    @testing.setter
    def testing(self, value):
        """Set the testing mode"""
        self.__dict__['testing'] = value and True or False

    @property
    def casting(self):
        """Check if the machine is casting"""
        return not self.punching and not self.testing

    @property
    def hmn(self):
        """Check if HMN mode is currently on"""
        return self.__dict__.get('_hmn', False)

    @hmn.setter
    def hmn(self, value):
        """Set the HMN mode"""
        value = value and True or False
        self.__dict__['_hmn'] = value

    @property
    def kmn(self):
        """Check if KMN mode is currently on"""
        return self.__dict__.get('_kmn', False)

    @kmn.setter
    def kmn(self, value):
        """Set the KMN mode"""
        value = value and True or False
        self.__dict__['_kmn'] = value

    @property
    def unitshift(self):
        """Check if the unit shift mode is currently on"""
        return self.__dict__.get('_unit-shift', False)

    @unitshift.setter
    def unitshift(self, value):
        """Set the unit-shift mode"""
        value = value and True or False
        self.__dict__['_unit-shift'] = value

    @property
    def sensor(self):
        """Chooses a proper sensor"""
        return (self.testing and TestSensor or
                self.punching and PunchingSensor or
                self.simulation and SimulationSensor or
                SENSORS.get(SENSOR, SimulationSensor))

    @property
    def output(self):
        """Chooses a simulation or hardware output driver"""
        return (self.simulation and SimulationOutput or
                OUTPUTS.get(OUTPUT, SimulationOutput))

    def choose_row16_addressing(self):
        """Let user decide which way to address row 16"""
        def hmn_on():
            """Turn on the HMN 16x17 attachment"""
            UI.display('Turn the HMN attachment ON')
            self.hmn = True

        def kmn_on():
            """Turn on the rare KMN 16x17 attachment"""
            UI.display('Turn the KMN attachment ON')
            self.kmn = True

        def unitshift_on():
            """Turn on the unit-shift attachment (introduced in 1960s)"""
            UI.display('Turn the attachment ON - switch under the paper tower')
            self.unitshift = True

        def do_nothing():
            """User can choose not to use any attachment"""
            return None

        prompt = ('Your ribbon contains codes from the 16th row.\n'
                  'It is supported by special attachments for the machine.\n'
                  'Which mode does your caster use: HMN, KMN, Unit-Shift?\n\n'
                  'If none - characters from row 15 will be cast instead.\n\n'
                  'Your choice: [U]nit-Shift, [H]MN, [K]MN '
                  '(leave blank for none)?: ')
        options = {'U': unitshift_on,
                   'H': hmn_on,
                   'K': kmn_on,
                   '': do_nothing}
        UI.simple_menu(prompt, options)()


def sysfs_sensor():
    """Gets hardware sensor - prevents import loop"""
    from .input_driver_sysfs import SysfsSensor
    return SysfsSensor()


def wiringpi_output():
    """Gets hardware output - prevents import loop"""
    from .output_driver_wiringpi import WiringPiOutputDriver
    return WiringPiOutputDriver()


def parallel_sensor():
    """A parallel port sensor for John Cornelisse's old interface"""
    pass


def parallel_output():
    """A parallel port valve control for John Cornelisse's old interface"""
    pass


SENSORS = {'simulation': SimulationSensor, 'sysfs': sysfs_sensor,
           'parallel': parallel_sensor}
OUTPUTS = {'simulation': SimulationOutput, 'wiringpi': wiringpi_output,
           'parallel': parallel_output}



from setuptools import setup, find_packages

with open('README.rst', 'r') as readme_file:
    long_description = readme_file.read()

clas = ['Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Manufacturing',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Artistic Software']

setup(name='rpi2caster',
      version='0.4.dev34',
      description='Raspberry Pi controls a Monotype composition caster',
      long_description=long_description,
      url='http://github.com/elegantandrogyne/rpi2caster',
      author='Christophe Slychan',
      author_email='krzysztof.slychan@gmail.com',
      license='GPLv3',
      zip_safe=True,
      include_package_data=True,
      classifiers=clas,
      extras_require={'output-driver-wiringpi': ['wiringpi >= 2.3'],
                      'output-driver-gpiozero': ['gpiozero >= 1.0'],
                      'input-driver-sysfs': ['gpiodaemon >= 0.5'],
                      'input-driver-gpiozero': ['gpiozero >= 1.0'],
                      'database-postgresql': ['psycopg2 >= 2.6']},
      install_requires=['pip > 1.5', 'click > 6.0'],
      keywords=['Monotype', 'typography', 'printing', 'letterpress',
                'typesetting', 'typecasting', 'Raspberry Pi'],
      packages=find_packages(exclude=['data', 'docs', 'tests']),
      entry_points={'console_scripts':
                    ['rpi2caster = rpi2caster.__main__:main']})



# -*- coding: utf-8 -*-
"""Typesetter program"""

import io
from . import exceptions as e
from . import typesetting_funcs as tsf
from .measure import Measure
from .typesetting_data import Ribbon
from .matrix_data import Diecase, diecase_operations
from .wedge_data import Wedge
from .global_settings import UI


class Typesetting(object):
    """Typesetting session - choose and translate text with control codes
    into a sequence of Monotype control codes, which can be sent to
    the machine to cast composed and justified type.
    """
    def __init__(self, text_file='', ribbon_file='', diecase_id='',
                 manual_mode=False):

        self.measure = Measure(manual=False)
        self.ribbon = Ribbon(ribbon_file)
        self.diecase = Diecase(diecase_id)
        self.source = text_file and open_file(text_file) or []
        # Use a manual compositor (user decides where to break the line)
        # or automatic compositor (hyphenation and justification is done
        # automatically with the Knuth-Plass algorithm)
        self.compositor = (manual_mode and ManualCompositor(self) or
                           AutoCompositor(self))

    def main_menu(self):
        """Main menu for the typesetting utility."""
        def finish():
            """Stop the loop"""
            nonlocal finished
            finished = True

        def choose_diecase():
            """Chooses a diecase from database"""
            self.diecase = Diecase(manual_choice=True)

        def choose_wedge():
            """Chooses a wedge from registered ones"""
            self.diecase.alt_wedge = Wedge(manual_choice=True)

        def change_measure():
            """Change a line length"""
            UI.display('Set the galley width...')
            self.measure = Measure()

        def menu_options():
            """Build a list of options, adding an option"""
            # Options are described with tuples:
            # (function, description, condition)
            opts = [(finish, 'Exit', 'Exits the program'),
                    (choose_diecase, 'Select diecase',
                     'Select a matrix case from database (current: %s)'
                     % (self.diecase or 'not selected')),
                    (choose_wedge, 'Select wedge',
                     'Enter a wedge designation (current: %s)'
                     % self.diecase.alt_wedge),
                    (change_measure, 'Change measure',
                     'Set new line length (current: %s)' % self.measure),
                    (self.diecase.show_layout, 'Show diecase layout',
                     'View the matrix case layout'),
                    (diecase_operations, 'Matrix manipulation',
                     'Work on matrix cases')]
            # Built a list of menu options conditionally
            return [(function, description, long_description)
                    for (function, description, long_description) in opts]

        header = ('rpi2caster - CAT (Computer-Aided Typecasting) '
                  'for Monotype Composition or Type and Rule casters.\n\n'
                  'Composition Menu:')
        # Keep displaying the menu and go back here after any method ends
        finished = False
        while not finished:
            # Catch any known exceptions here
            try:
                UI.menu(menu_options(), header=header, footer='')()
            except (e.ReturnToMenu, e.MenuLevelUp, KeyboardInterrupt):
                # Will skip to the end of the loop, and start all over
                pass


class Compositor(object):
    """Composition engine class"""
    def __init__(self, context):
        self.context = context


class ManualCompositor(Compositor):
    """Manual composition - allows more control over typesetting process"""
    def __init__(self, context):
        super().__init__(context)


class AutoCompositor(Compositor):
    """Automatic composition with hyphenation and justification"""
    def __init__(self, context):
        super().__init__(context)


def open_file(filename=''):
    """Opens a text file with text that will be typeset"""
    while True:
        # Choose file
        try:
            filename = filename or UI.enter_input_filename()
        except e.ReturnToMenu:
            return []
        # Open it
        with io.open(filename, 'r') as text_file:
            return text_file.readlines()
